@@ -23,11 +23,11 @@ const ENTERED_MENU_COLOR = new Clutter.Color();
 ENTERED_MENU_COLOR.from_pixel(0x00ff0022);
 
 const WELL_DEFAULT_COLUMNS = 4;
-const WELL_ITEM_HSPACING = 0;
+const WELL_ITEM_MIN_HSPACING = 4;
 const WELL_ITEM_VSPACING = 4;
 
-const MENU_ICON_SIZE = 24;
-const MENU_SPACING = 15;
+const MENU_ARROW_SIZE = 12;
+const MENU_SPACING = 7;
 
 const MAX_ITEMS = 30;
 
@@ -57,7 +57,13 @@ AppDisplayItem.prototype = {
 
     // Opens an application represented by this display item.
     launch : function() {
-        this._appInfo.launch();
+        let windows = Shell.AppMonitor.get_default().get_windows_for_app(this._appInfo.get_id());
+        if (windows.length > 0) {
+            let mostRecentWindow = windows[0];
+            Main.overview.activateWindow(mostRecentWindow, Main.currentTime());
+        } else {
+            this._appInfo.launch();
+        }
     },
 
     //// Protected method overrides ////
@@ -81,8 +87,8 @@ const MENU_UNSELECTED = 0;
 const MENU_SELECTED = 1;
 const MENU_ENTERED = 2;
 
-function MenuItem(name, id, iconName) {
-    this._init(name, id, iconName);
+function MenuItem(name, id) {
+    this._init(name, id);
 }
 
 /**
@@ -90,31 +96,19 @@ function MenuItem(name, id, iconName) {
  * Shows the list of menus in the sidebar.
  */
 MenuItem.prototype = {
-    _init: function(name, id, iconName) {
+    _init: function(name, id) {
         this.id = id;
 
         this.actor = new Big.Box({ orientation: Big.BoxOrientation.HORIZONTAL,
                                    spacing: 4,
                                    corner_radius: 4,
                                    padding_right: 4,
+                                   padding_left: 4,
                                    reactive: true });
         this.actor.connect('button-press-event', Lang.bind(this, function (a, e) {
             this.setState(MENU_SELECTED);
         }));
 
-        let iconTheme = Gtk.IconTheme.get_default();
-        let pixbuf = null;
-        this._icon = new Clutter.Texture({ width: MENU_ICON_SIZE,
-                                           height: MENU_ICON_SIZE });
-        // Wine manages not to have an icon
-        try {
-            pixbuf = iconTheme.load_icon(iconName, MENU_ICON_SIZE, 0 /* flags */);
-        } catch (e) {
-            pixbuf = iconTheme.load_icon('gtk-file', MENU_ICON_SIZE, 0);
-        }
-        if (pixbuf != null)
-            Shell.clutter_texture_set_from_pixbuf(this._icon, pixbuf);
-        this.actor.append(this._icon, Big.BoxPackFlags.NONE);
         this._text = new Clutter.Text({ color: GenericDisplay.ITEM_DISPLAY_NAME_COLOR,
                                         font_name: "Sans 14px",
                                         text: name });
@@ -122,7 +116,8 @@ MenuItem.prototype = {
         // We use individual boxes for the label and the arrow to ensure that they
         // are aligned vertically. Just setting y_align: Big.BoxAlignment.CENTER
         // on this.actor does not seem to achieve that.  
-        let labelBox = new Big.Box({ y_align: Big.BoxAlignment.CENTER });
+        let labelBox = new Big.Box({ y_align: Big.BoxAlignment.CENTER,
+                                     padding: 4 });
 
         labelBox.append(this._text, Big.BoxPackFlags.NONE);
        
@@ -130,8 +125,8 @@ MenuItem.prototype = {
 
         let arrowBox = new Big.Box({ y_align: Big.BoxAlignment.CENTER });
 
-        this._arrow = new Shell.Arrow({ surface_width: MENU_ICON_SIZE / 2,
-                                        surface_height: MENU_ICON_SIZE / 2,
+        this._arrow = new Shell.Arrow({ surface_width: MENU_ARROW_SIZE,
+                                        surface_height: MENU_ARROW_SIZE,
                                         direction: Gtk.ArrowType.RIGHT,
                                         opacity: 0 });
         arrowBox.append(this._arrow, Big.BoxPackFlags.NONE);
@@ -161,48 +156,42 @@ MenuItem.prototype = {
 }
 Signals.addSignalMethods(MenuItem.prototype);
 
-
 /* This class represents a display containing a collection of application items.
  * The applications are sorted based on their popularity by default, and based on
  * their name if some search filter is applied.
+ *
+ * showPrefs - a boolean indicating if this AppDisplay should contain preference
+ *             applets, rather than applications
  */
-function AppDisplay() {
-    this._init();
+function AppDisplay(showPrefs) {
+    this._init(showPrefs);
 }
 
 AppDisplay.prototype = {
     __proto__:  GenericDisplay.GenericDisplay.prototype,
 
-    _init : function() {
+    _init : function(showPrefs) {
         GenericDisplay.GenericDisplay.prototype._init.call(this);
 
+        this._showPrefs = showPrefs;
+
         this._menus = [];
         this._menuDisplays = [];
-
-        // map<itemId, array of category names>
-        this._appCategories = {};
+        // map<search term, map<appId, true>>
+        // We use a map of appIds instead of an array to ensure that we don't have duplicates and for easier lookup.
+        this._menuSearchAppMatches = {};
 
         this._appMonitor = Shell.AppMonitor.get_default();
         this._appSystem = Shell.AppSystem.get_default();
         this._appsStale = true;
         this._appSystem.connect('installed-changed', Lang.bind(this, function(appSys) {
             this._appsStale = true;
-            this._redisplay(false);
-            this._redisplayMenus();
+            this._redisplay(GenericDisplay.RedisplayFlags.NONE);
         }));
         this._appSystem.connect('favorites-changed', Lang.bind(this, function(appSys) {
-            this._redisplay(false);
+            this._appsStale = true;
+            this._redisplay(GenericDisplay.RedisplayFlags.NONE);
         }));
-        this._appMonitor.connect('app-added', Lang.bind(this, function(monitor) {
-            this._redisplay(false);
-        }));
-        this._appMonitor.connect('app-removed', Lang.bind(this, function(monitor) {
-            this._redisplay(false);
-        }));
-
-        // Load the apps now so it doesn't slow down the first
-        // transition into the Overview
-        this._refreshCache();
 
         this._focusInMenus = true;
         this._activeMenuIndex = -1;
@@ -211,7 +200,6 @@ AppDisplay.prototype = {
         this._menuDisplay = new Big.Box({ orientation: Big.BoxOrientation.VERTICAL,
                                           spacing: MENU_SPACING
                                        });
-        this._redisplayMenus();
 
         this.connect('expanded', Lang.bind(this, function (self) {
             this._filterReset();
@@ -254,6 +242,39 @@ AppDisplay.prototype = {
         return true;
     },
 
+    setSearch: function(text) {
+        let lowertext = text.toLowerCase();
+        if (lowertext == this._search)
+            return;
+
+        // We prepare menu matches up-front, so that we don't
+        // need to go over all menu items for each application
+        // and then get all applications for a matching menu
+        // to see if a particular application passed to
+        // _isInfoMatching() is a match.
+        let terms = lowertext.split(/\s+/);
+        this._menuSearchAppMatches = {};
+        for (let i = 0; i < terms.length; i++) {
+            let term = terms[i];
+            this._menuSearchAppMatches[term] = {};
+            for (let j = 0; j < this._menus.length; j++) {
+                let menuItem = this._menus[j];
+                // Match only on the beginning of the words in category names,
+                // because otherwise it introduces unnecessary noise in the results.
+                if (menuItem.name.toLowerCase().indexOf(term) == 0 ||
+                    menuItem.name.toLowerCase().indexOf(" " + term) > 0) {
+                    let menuApps = this._appSystem.get_applications_for_menu(menuItem.id);
+                    for (let k = 0; k < menuApps.length; k++) {
+                        let menuApp = menuApps[k];
+                        this._menuSearchAppMatches[term][menuApp.get_id()] = true;
+                    }
+                }
+            }
+        }
+
+        GenericDisplay.GenericDisplay.prototype.setSearch.call(this, text);
+    },
+
     // Protected overrides
 
     _filterActive: function() {
@@ -285,8 +306,8 @@ AppDisplay.prototype = {
         })).filter(function (e) { return e != null });
     },
 
-    _addMenuItem: function(name, id, icon, index) {
-        let display = new MenuItem(name, id, icon);
+    _addMenuItem: function(name, id, index) {
+        let display = new MenuItem(name, id);
         this._menuDisplays.push(display);
         display.connect('state-changed', Lang.bind(this, function (display) {
             let activated = display.getState() != MENU_UNSELECTED;
@@ -304,7 +325,7 @@ AppDisplay.prototype = {
                     this._activeMenuApps = this._appSystem.get_applications_for_menu(id);
                 }
             }
-            this._redisplay(true);
+            this._redisplay(GenericDisplay.RedisplayFlags.FULL);
         }));
         this._menuDisplay.append(display.actor, 0);
     },
@@ -312,9 +333,13 @@ AppDisplay.prototype = {
     _redisplayMenus: function() {
         this._menuDisplay.remove_all();
         this._addMenuItem(_("Frequent"), null, 'gtk-select-all');
+        // Adding an empty box here results in double spacing between
+        // "Frequent" and the other items.
+        let separator_actor = new Big.Box();
+        this._menuDisplay.append(separator_actor, 0);
         for (let i = 0; i < this._menus.length; i++) {
             let menu = this._menus[i];
-            this._addMenuItem(menu.name, menu.id, menu.icon, i+1);
+            this._addMenuItem(menu.name, menu.id, i+1);
         }
     },
 
@@ -327,44 +352,47 @@ AppDisplay.prototype = {
 
     // Gets information about all applications by calling Gio.app_info_get_all().
     _refreshCache : function() {
-        let me = this;
         if (!this._appsStale)
-            return;
+            return true;
         this._allItems = {};
-        this._appCategories = {};
 
-        // Loop over the toplevel menu items, load the set of desktop file ids
-        // associated with each one, skipping empty menus
-        let allMenus = this._appSystem.get_menus();
-        this._menus = [];
-        for (let i = 0; i < allMenus.length; i++) {
-            let menu = allMenus[i];
-            let menuApps = this._appSystem.get_applications_for_menu(menu.id);
-            let hasVisibleApps = menuApps.some(function (app) { return !app.get_is_nodisplay(); });
-            if (!hasVisibleApps) {
-                continue;
-            }
-            this._menus.push(menu);
-            for (let j = 0; j < menuApps.length; j++) {
-                let app = menuApps[j];
+        if (this._showPrefs) {
+            // Get the desktop file ids for settings/preferences.
+            // These are used for search results, but not in the app menus.
+            let settings = this._appSystem.get_all_settings();
+            for (let i = 0; i < settings.length; i++) {
+                let app = settings[i];
                 this._addApp(app);
             }
-        }
-
-        // Now grab the desktop file ids for settings/preferences.
-        // These show up in search, but not with the rest of apps.
-        let settings = this._appSystem.get_all_settings();
-        for (let i = 0; i < settings.length; i++) {
-            let app = settings[i];
-            this._addApp(app);
+        } else {
+            // Loop over the toplevel menu items, load the set of desktop file ids
+            // associated with each one, skipping empty menus
+            let allMenus = this._appSystem.get_menus();
+            this._menus = [];
+            for (let i = 0; i < allMenus.length; i++) {
+                let menu = allMenus[i];
+                let menuApps = this._appSystem.get_applications_for_menu(menu.id);
+                let hasVisibleApps = menuApps.some(function (app) { return !app.get_is_nodisplay(); });
+                if (!hasVisibleApps) {
+                    continue;
+                }
+                this._menus.push(menu);
+                for (let j = 0; j < menuApps.length; j++) {
+                    let app = menuApps[j];
+                    this._addApp(app);
+                }
+            }
+            this._redisplayMenus();
         }
 
         this._appsStale = false;
+        return false;
     },
 
     // Stub this out; the app display always has a category selected
     _setDefaultList : function() {
-        this._matchedItems = [];
+        this._matchedItems = {};
+        this._matchedItemKeys = [];
     },
 
     // Compares items associated with the item ids based on the alphabetical order
@@ -377,7 +405,7 @@ AppDisplay.prototype = {
     },
 
     // Checks if the item info can be a match for the search string by checking
-    // the name, description, execution command, and categories for the application.
+    // the name, description, execution command, and category for the application.
     // Item info is expected to be Shell.AppInfo.
     // Returns a boolean flag indicating if itemInfo is a match.
     _isInfoMatching : function(itemInfo, search) {
@@ -389,10 +417,10 @@ AppDisplay.prototype = {
         if (this._activeMenu == null || search != "")
             return this._isInfoMatchingSearch(itemInfo, search);
         else
-            return this._isInfoMatchingMenu(itemInfo, search);
+            return this._isInfoMatchingMenu(itemInfo);
     },
 
-    _isInfoMatchingMenu : function(itemInfo, search) {
+    _isInfoMatchingMenu: function(itemInfo) {
         let id = itemInfo.get_id();
         for (let i = 0; i < this._activeMenuApps.length; i++) {
             let activeApp = this._activeMenuApps[i];
@@ -430,11 +458,11 @@ AppDisplay.prototype = {
                 return true;
         }
 
-        let categories = itemInfo.get_categories();
-        for (let i = 0; i < categories.length; i++) {
-            let category = fold(categories[i]);
-            if (category.indexOf(search) >= 0)
+        if (this._menuSearchAppMatches[search]) {
+            if (this._menuSearchAppMatches[search].hasOwnProperty(itemInfo.get_id()))
                 return true;
+        } else {
+            log("Missing an entry for search term " + search + " in this._menuSearchAppMatches");
         }
 
         return false;
@@ -448,35 +476,42 @@ AppDisplay.prototype = {
 
 Signals.addSignalMethods(AppDisplay.prototype);
 
-function WellDisplayItem(appInfo, isFavorite) {
-    this._init(appInfo, isFavorite);
+function BaseWellItem(appInfo, isFavorite, hasMenu) {
+    this._init(appInfo, isFavorite, hasMenu);
 }
 
-WellDisplayItem.prototype = {
-    __proto__ : AppIcon.AppIcon.prototype,
+BaseWellItem.prototype = {
+    __proto__: AppIcon.AppIcon.prototype,
 
-    _init : function(appInfo, isFavorite) {
-        AppIcon.AppIcon.prototype._init.call(this, appInfo);
+    _init: function(appInfo, isFavorite) {
+        AppIcon.AppIcon.prototype._init.call(this, { appInfo: appInfo,
+                                                     menuType: AppIcon.MenuType.ON_RIGHT,
+                                                     glow: true });
 
         this.isFavorite = isFavorite;
 
-        this.actor.connect('button-release-event', Lang.bind(this, function (b, e) {
-            this._handleActivate();
+        this._draggable = DND.makeDraggable(this.actor, true);
+
+        // Do these as anonymous functions to avoid conflict with handlers in subclasses
+        this.actor.connect('button-press-event', Lang.bind(this, function(actor, event) {
+            let [stageX, stageY] = event.get_coords();
+            this._dragStartX = stageX;
+            this._dragStartY = stageY;
+            return false;
+        }));
+        this.actor.connect('notify::hover', Lang.bind(this, function () {
+            let hover = this.actor.hover;
+            if (!hover) {
+                if (this.actor.pressed && this._dragStartX != null) {
+                    this.actor.fake_release();
+                    this._draggable.startDrag(this._dragStartX, this._dragStartY,
+                                              Main.currentTime());
+                } else {
+                    this._dragStartX = null;
+                    this._dragStartY = null;
+                }
+            }
         }));
-
-        let draggable = DND.makeDraggable(this.actor);
-    },
-
-    _handleActivate: function () {
-        if (this._windows.length == 0) {
-            this.appInfo.launch();
-            Main.overview.hide();
-        } else {
-            /* Pick the first window and activate it;
-             * In the future, we want to have a menu dropdown here. */
-            let first = this._windows[0];
-            Main.overview.activateWindow(first, Clutter.get_current_event_time());
-        }
     },
 
     shellWorkspaceLaunch : function() {
@@ -491,18 +526,104 @@ WellDisplayItem.prototype = {
         this.appInfo.launch();
     },
 
-    getDragActor: function(stageX, stageY) {
-        return this.appInfo.create_icon_texture(this._icon.height);
+    getDragActor: function() {
+        return this.createDragActor();
     },
 
     // Returns the original icon that is being used as a source for the cloned texture
     // that represents the item as it is being dragged.
     getDragActorSource: function() {
-        return this._icon;
+        return this.actor;
+    }
+}
+
+function RunningWellItem(appInfo, isFavorite) {
+    this._init(appInfo, isFavorite);
+}
+
+RunningWellItem.prototype = {
+    __proto__: BaseWellItem.prototype,
+
+    _init: function(appInfo, isFavorite) {
+        BaseWellItem.prototype._init.call(this, appInfo, isFavorite);
+
+        this._dragStartX = 0;
+        this._dragStartY = 0;
+
+        this.actor.connect('activate', Lang.bind(this, this._onActivate));
     },
 
-    setWidth: function(width) {
-        this._nameBox.width = width + GLOW_PADDING * 2;
+    _onActivate: function (actor, event) {
+        let modifiers = Shell.get_event_state(event);
+
+        if (modifiers & Clutter.ModifierType.CONTROL_MASK) {
+            this.appInfo.launch();
+        } else {
+            this.activateMostRecentWindow();
+        }
+    },
+
+    activateMostRecentWindow: function () {
+        // The _get_windows_for_app sorts them for us
+        let mostRecentWindow = this.windows[0];
+        Main.overview.activateWindow(mostRecentWindow, Main.currentTime());
+    },
+
+    highlightWindow: function(metaWindow) {
+        Main.overview.getWorkspacesForWindow(metaWindow).setHighlightWindow(metaWindow);
+    },
+
+    activateWindow: function(metaWindow) {
+        if (metaWindow) {
+            this._didActivateWindow = true;
+            Main.overview.activateWindow(metaWindow, Main.currentTime());
+        } else
+            Main.overview.hide();
+    },
+
+    menuPoppedUp: function() {
+        Main.overview.getWorkspacesForWindow(null).setApplicationWindowSelection(this.appInfo.get_id());
+    },
+
+    menuPoppedDown: function() {
+        if (this._didActivateWindow)
+            return;
+
+        Main.overview.getWorkspacesForWindow(null).setApplicationWindowSelection(null);
+    }
+};
+
+function InactiveWellItem(appInfo, isFavorite) {
+    this._init(appInfo, isFavorite);
+}
+
+InactiveWellItem.prototype = {
+    __proto__: BaseWellItem.prototype,
+
+    _init : function(appInfo, isFavorite) {
+        BaseWellItem.prototype._init.call(this, appInfo, isFavorite);
+
+        this.actor.connect('notify::pressed', Lang.bind(this, this._onPressedChanged));
+        this.actor.connect('activate', Lang.bind(this, this._onActivate));
+    },
+
+    _onPressedChanged: function() {
+        this.setHighlight(this.actor.pressed);
+    },
+
+    _onActivate: function() {
+        if (this.windows.length == 0) {
+            this.appInfo.launch();
+            Main.overview.hide();
+            return true;
+        }
+        return false;
+    },
+
+    menuPoppedUp: function() {
+    },
+
+    menuPoppedDown: function() {
     }
 };
 
@@ -532,9 +653,8 @@ WellGrid.prototype = {
             nColumns = children.length;
         else
             nColumns = WELL_DEFAULT_COLUMNS;
-        let spacing = Math.max(nColumns - 1, 0) * WELL_ITEM_HSPACING;
-        alloc.min_size = itemMin * nColumns + spacing;
-        alloc.natural_size = itemNatural * nColumns + spacing;
+        alloc.min_size = itemMin;
+        alloc.natural_size = itemNatural * nColumns;
     },
 
     _getPreferredHeight: function (grid, forWidth, alloc) {
@@ -571,10 +691,8 @@ WellGrid.prototype = {
             childBox.y2 = childBox.y1 + itemHeight;
             children[i].allocate(childBox, flags);
 
-            let atSeparator = (i == this._separatorIndex - 1);
-
             columnIndex++;
-            if (columnIndex == columns || atSeparator) {
+            if (columnIndex == columns) {
                 columnIndex = 0;
             }
 
@@ -582,29 +700,9 @@ WellGrid.prototype = {
                 y += itemHeight + WELL_ITEM_VSPACING;
                 x = box.x1;
             } else {
-                x += itemWidth + WELL_ITEM_HSPACING;
-            }
-
-            if (atSeparator) {
-                y += separatorNatural + WELL_ITEM_VSPACING;
+                x += itemWidth;
             }
         }
-
-        let separatorRowIndex = Math.ceil(this._separatorIndex / columns);
-
-        /* Allocate the separator */
-        let childBox = new Clutter.ActorBox();
-        childBox.x1 = box.x1;
-        childBox.y1 = (itemHeight + WELL_ITEM_VSPACING) * separatorRowIndex;
-        this._cachedSeparatorY = childBox.y1;
-        childBox.x2 = box.x2;
-        childBox.y2 = childBox.y1+separatorNatural;
-        this._separator.allocate(childBox, flags);
-    },
-
-    setSeparatorIndex: function (index) {
-        this._separatorIndex = index;
-        this.actor.queue_relayout();
     },
 
     removeAll: function () {
@@ -612,11 +710,6 @@ WellGrid.prototype = {
         for (let i = 0; i < itemChildren.length; i++) {
             itemChildren[i].destroy();
         }
-        this._separatorIndex = 0;
-    },
-
-    isBeforeSeparator: function(x, y) {
-        return y < this._cachedSeparatorY;
     },
 
     _getItemChildren: function () {
@@ -632,29 +725,40 @@ WellGrid.prototype = {
         let children = this._getItemChildren();
         if (children.length == 0)
             return [0, WELL_DEFAULT_COLUMNS, 0, 0];
-        let nColumns;
-        if (children.length < WELL_DEFAULT_COLUMNS)
-            nColumns = children.length;
-         else
+        let nColumns = 0;
+        let usedWidth = 0;
+        // Big.Box will allocate us at 0x0 if we are not visible; this is probably a
+        // Big.Box bug but it can't be fixed because if children are skipped in allocate()
+        // Clutter gets confused (see http://bugzilla.openedhand.com/show_bug.cgi?id=1831)
+        if (forWidth <= 0) {
             nColumns = WELL_DEFAULT_COLUMNS;
-
-        if (forWidth >= 0 && forWidth < minWidth) {
-           log("WellGrid: trying to allocate for width " + forWidth + " but min is " + minWidth);
-           /* FIXME - we should fall back to fewer than WELL_DEFAULT_COLUMNS here */
+        } else {
+            while (nColumns < WELL_DEFAULT_COLUMNS &&
+                   nColumns < children.length &&
+                   usedWidth + itemMinWidth <= forWidth) {
+               // By including WELL_ITEM_MIN_HSPACING in usedWidth, we are ensuring
+               // that the number of columns we end up with will allow the spacing
+               // between the columns to be at least that value.
+               usedWidth += itemMinWidth + WELL_ITEM_MIN_HSPACING;
+               nColumns++;
+            }
         }
 
-        let horizSpacingTotal = Math.max(nColumns - 1, 0) * WELL_ITEM_HSPACING;
-        let minWidth = itemMinWidth * nColumns + horizSpacingTotal;
+        if (nColumns == 0) {
+           log("WellGrid: couldn't fit a column in width " + forWidth);
+           /* FIXME - fall back to smaller icon size */
+        }
+
+        let minWidth = itemMinWidth * nColumns;
 
         let lastColumnIndex = nColumns - 1;
-        let separatorColumns = lastColumnIndex - ((lastColumnIndex + this._separatorIndex) % nColumns);
-        let rows = Math.ceil((children.length + separatorColumns) / nColumns);
+        let rows = Math.ceil(children.length / nColumns);
 
         let itemWidth;
-        if (forWidth < 0) {
+        if (forWidth <= 0) {
             itemWidth = itemNaturalWidth;
         } else {
-            itemWidth = Math.max(forWidth - horizSpacingTotal, 0) / nColumns;
+            itemWidth = Math.floor(forWidth / nColumns);
         }
 
         let itemNaturalHeight = 0;
@@ -664,7 +768,7 @@ WellGrid.prototype = {
                 itemNaturalHeight = childNatural;
         }
 
-        return [rows, WELL_DEFAULT_COLUMNS, itemWidth, itemNaturalHeight];
+        return [rows, nColumns, itemWidth, itemNaturalHeight];
     },
 
     _getItemPreferredWidth: function () {
@@ -695,6 +799,9 @@ AppWell.prototype = {
                                    x_align: Big.BoxAlignment.CENTER });
         this.actor._delegate = this;
 
+        this._pendingRedisplay = false;
+        this.actor.connect('notify::mapped', Lang.bind(this, this._onMappedNotify));
+
         this._grid = new WellGrid();
         this.actor.append(this._grid.actor, Big.BoxPackFlags.EXPAND);
 
@@ -734,7 +841,20 @@ AppWell.prototype = {
                             values[id] = index; return values; }, {});
     },
 
+    _onMappedNotify: function() {
+        let mapped = this.actor.mapped;
+        if (mapped && this._pendingRedisplay)
+            this._redisplay();
+    },
+
     _redisplay: function () {
+        let mapped = this.actor.mapped;
+        if (!mapped) {
+            this._pendingRedisplay = true;
+            return;
+        }
+        this._pendingRedisplay = false;
+
         this._grid.removeAll();
 
         let favoriteIds = this._appSystem.get_favorites();
@@ -750,29 +870,29 @@ AppWell.prototype = {
 
         let displays = []
         this._addApps(favorites, true);
-        this._grid.setSeparatorIndex(favorites.length);
         this._addApps(running, false);
         this._displays = displays;
     },
 
-    _addApps: function(apps) {
+    _addApps: function(apps, isFavorite) {
         for (let i = 0; i < apps.length; i++) {
             let app = apps[i];
-            let display = new WellDisplayItem(app, this.isFavorite);
+            let windows = this._appMonitor.get_windows_for_app(app.get_id());
+            let display;
+            if (windows.length > 0)
+                display = new RunningWellItem(app, isFavorite);
+            else
+                display = new InactiveWellItem(app, isFavorite);
             this._grid.actor.add_actor(display.actor);
         }
     },
 
     // Draggable target interface
     acceptDrop : function(source, actor, x, y, time) {
-        let global = Shell.Global.get();
-
         let appSystem = Shell.AppSystem.get_default();
 
         let app = null;
-        if (source instanceof WellDisplayItem) {
-            app = source.appInfo;
-        } else if (source instanceof AppDisplayItem) {
+        if (source instanceof AppDisplayItem) {
             app = appSystem.lookup_cached_app(source.getId());
         } else if (source instanceof Workspaces.WindowClone) {
             let appMonitor = Shell.AppMonitor.get_default();
@@ -789,22 +909,15 @@ AppWell.prototype = {
         let favoriteIds = this._appSystem.get_favorites();
         let favoriteIdsObject = this._arrayValues(favoriteIds);
 
-        let dropIsFavorite = this._grid.isBeforeSeparator(x - this._grid.actor.x,
-                                                          y - this._grid.actor.y);
         let srcIsFavorite = (id in favoriteIdsObject);
 
-        if (srcIsFavorite && (!dropIsFavorite)) {
-            Mainloop.idle_add(function () {
-                appSystem.remove_favorite(id);
-                return false;
-            });
-        } else if ((!srcIsFavorite) && dropIsFavorite) {
+        if (srcIsFavorite) {
+            return false;
+        } else {
             Mainloop.idle_add(function () {
                 appSystem.add_favorite(id);
                 return false;
             });
-        } else {
-            return false;
         }
 
         return true;
